import datetime
import os
import pathlib
import tempfile
import unittest

from typing import Type
from os.path import join
from os import environ

import yaml

from ninja_bear import (
    DumpInfo,
    GeneratorBase,
    LanguageConfigBase,
    NamingConventionType,
    Orchestrator,
    Plugin,
    PropertyType,
)
from src.ninja_bear_distributor_git.distributor import Distributor, execute_command


class ExampleScriptGenerator(GeneratorBase):
    """
    ExampleScript specific generator. For more information about the generator methods, refer to GeneratorBase.
    """

    def _default_type_naming_convention(self) -> NamingConventionType:
        return NamingConventionType.PASCAL_CASE
    
    def _line_comment(self, string: str) -> str:
        return f'-- {string}'
    
    def _dump(self, info: DumpInfo) -> str:
        code = f'struct {info.type_name}:\n'

        for property in info.properties:
            type = property.type
            value = property.value

            if type == PropertyType.BOOL:
                type_string = 'boolean'
                value = 'true' if value else 'false'
            elif type == PropertyType.INT:
                type_string = 'int'
            elif type == PropertyType.FLOAT:
                type_string = 'float'
            elif type == PropertyType.DOUBLE:
                type_string = 'double'
            elif type == PropertyType.STRING:
                type_string = 'string'
                value = f'\'{value}\''
            elif type == PropertyType.REGEX:
                type_string = 'regex'
                value = f'/{value}/'

            comment = f' {self._line_comment(property.comment)}' if property.comment else ''
            code += f'{" " * info.indent}{type_string} {property.name} = {value}{comment}\n'

        return code


class ExampleScriptConfig(LanguageConfigBase):
    """
    ExampleScript specific config. For more information about the config methods, refer to LanguageConfigBase.
    """

    def _file_extension(self) -> str:
        return 'es'

    def _generator_type(self) -> Type[ExampleScriptGenerator]:
        return ExampleScriptGenerator
    
    def _default_file_naming_convention(self) -> NamingConventionType:
        return NamingConventionType.KEBAP_CASE

    def _allowed_file_name_pattern(self) -> str:
        return r'.+'


class Test(unittest.TestCase):
    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)
        self._config_name = 'test-config'
        self._config_file = f'{self._config_name}.yaml'
        self._test_path = pathlib.Path(__file__).parent.resolve()
        self._test_config_path = join(self._test_path, '..', f'example/{self._config_file}')
        self._plugins = [
            Plugin('examplescript', ExampleScriptConfig),
            Plugin('ninja-bear-distributor-git', Distributor),
        ]

    def test_distribution(self):
        def distribute(change: bool):
            # Load test-config.yaml directly in test file to allow implementer to modify properties if required.
            with open(self._test_config_path, 'r') as f:
                config = yaml.safe_load(f)
                remote = environ.get('URL')
                token = environ.get('TOKEN')
                start_datetime = datetime.datetime.now()
                commit_message = f'Updating config ({start_datetime.isoformat()})'

                if not remote:
                    raise Exception('No remote URL provided')
                if not token:
                    raise Exception('No authentication token provided')
                
                # Update data in distributor.
                git_distributor = config['distributors'][0]
                git_distributor['url'] = remote
                git_distributor['password'] = token
                git_distributor['message'] = commit_message
                del git_distributor['user']

                # If change shall be tested, change float property.
                if change:
                    config['properties'][2]['value'] = start_datetime.timestamp()

                # If context is a CI server, set the Git user.
                if os.getenv('GITHUB_ACTIONS') or os.getenv('TRAVIS') or \
                    os.getenv('CIRCLECI') or os.getenv('GITLAB_CI'):
                    execute_command('git config --global user.name github-actions')
                    execute_command('git config --global user.email github-actions@github.com')
                
                # Run parsing and distribution.
                orchestrator = Orchestrator.parse_config(config, self._config_name, plugins=self._plugins)
                orchestrator.distribute()

                with tempfile.TemporaryDirectory() as temp_dir:
                    code, _, _ = execute_command(f' git clone {remote.replace("://", f"://{token}@")} {temp_dir}')
                    
                    if code != 0:
                        raise Exception('Cloning failed')

                    _, compare_commit_message, _ = execute_command('git log -1 --pretty=%B', directory=temp_dir)  # https://stackoverflow.com/a/7293026

                    if change:
                        self.assertEqual(commit_message.strip(), compare_commit_message.strip())

        # First call to check if upload works.
        distribute(True)

        # Second call to undo value change.
        distribute(False)

        # Third call to check 'nothing to commit'-branch.
        distribute(False)



from __future__ import annotations
import os
import re
import subprocess
import tempfile

from typing import Dict, Tuple
from getpass import getpass
from os.path import join

from ninja_bear import DistributorBase, DistributorCredentials, DistributeInfo

class NoRepositoryUrlProvidedException(Exception):
    def __init__(self):
        super().__init__('No repository URL has been provided')


class NoCommitMessageProvidedException(Exception):
    def __init__(self, file_name, url):
        super().__init__(f'No commit message has been provided for {file_name} ({url})')


class GitProblemException(Exception):
    def __init__(self, problem: str, additional_info: str=''):
        problem = f'{problem}\n{additional_info}' if additional_info else problem
        super().__init__(problem)


class GitVersionException(Exception):
    def __init__(self, check_version: GitVersion, git_version: GitVersion):
        super().__init__(f'Invalid git version. Required: {check_version}, actual: {git_version}')


def execute_command(commands: str, directory=None) -> Tuple[int, str, str]:
        # https://www.squash.io/how-to-execute-a-program-or-system-command-in-python/.
        result = subprocess.run(commands, capture_output=True, text=True, shell=True, cwd=directory)
        return result.returncode, result.stdout, result.stderr


class GitVersion:
    major: str
    minor: str
    patch: str

    def __init__(self, major: int, minor: int, patch: int):
        """
        Constructor

        :param major: Major version.
        :type major:  int
        :param minor: Minor version.
        :type minor:  int
        :param patch: Patch version.
        :type patch:  int
        """
        self.major = major
        self.minor = minor
        self.patch = patch

    @staticmethod
    def from_git() -> GitVersion:
        """
        Retrieves the git version by calling git --version.

        :raises GitProblemException: Raised if the Git version could not be evaluated.

        :return: GitVersion instance with the actual Git version.
        :rtype:  GitVersion
        """
        code, text, _ = execute_command('git --version')
        version_text = text if code == 0 else ''

        if code != 0:
            raise GitProblemException('Git version could not be evaluated')
        return GitVersion.from_string(version_text)

    @staticmethod
    def from_string(version_string: str) -> GitVersion:
        """
        Turns a version string.

        :param version_string: Version string which contains the version in the form of <major>.<minor>.<patch>.
        :type version_string:  str

        :return: GitVersion instance with the data from the version string.
        :rtype:  GitVersion
        """
        version = GitVersion(0, 0, 0)
        match = re.search('\d+\.\d+\.\d+', version_string)

        if match:
            parts = match.group(0).split('.')

            version.major = int(parts[0])
            version.minor = int(parts[1])
            version.patch = int(parts[2])
        return version
    
    def __str__(self) -> str:
        return f'{self.major}.{self.minor}.{self.patch}'


class Distributor(DistributorBase):
    _MIN_GIT_VERSION = GitVersion(2, 29, 0)  # Needs at least git version 2.29.0 as it introduced partial-clone (https://www.git-scm.com/docs/partial-clone).

    def __init__(self, config: Dict[str, str], credentials: DistributorCredentials=None):
        """
        Constructor

        :param url:         Git repository URL.
        :type url:          str
        :param target_path: Relative target path within the repo.
        :type target_path:  str
        :param user:        Git user.
        :type user:         str
        :param password:    Git user password or token.
        :type password:     str
        """
        super().__init__(config, credentials)

        target_path, _ = self.from_config('path')  # Use root directory as default path.

        self._url, _ = self.from_config('url')
        self._target_path = target_path if target_path else ''  # Use root directory as default path.
        self._user, self._user_key_exists = self.from_config('user')
        self._password, self._password_key_exists = self.from_config('password')
        self._branch, _ = self.from_config('branch')
        self._message, _ = self.from_config('message')

        self._check_preconditions()
        self._check_git_version()
        
    def _check_preconditions(self):
        """
        Check preconditions.

        :raises NoRepositoryUrlProvidedException: Raised if no Git server URL has been provided.
        """
        url = self._url

        # Make sure an URL has been provided.
        if not url:
            raise NoRepositoryUrlProvidedException()
        
    def _check_git_version(self):
        """
        Checks if the installed Git version is suitable.

        :raises GitVersionException: Raised if Git does not fulfill the minimum required version.
        """
        # Retrieve Git version.
        git_version = GitVersion.from_git()
        
        if git_version.major < self._MIN_GIT_VERSION.major or \
           git_version.minor < self._MIN_GIT_VERSION.minor or \
           git_version.patch < self._MIN_GIT_VERSION.patch:
            raise GitVersionException(self._MIN_GIT_VERSION, git_version)

    def _distribute(self, info: DistributeInfo) -> DistributorBase:
        """
        Method to distribute a generated config to a Git server.

        :param info: Contains the required information to distribute the generated config.
        :type info:  DistributeInfo

        :raises GitProblemException: Raised on different Git problems.

        :return: The current GitDistributor instance.
        :rtype:  GitDistributor
        """
        file_name = info.file_name
        data = info.data

        # Create temporary folder to clone the git repo into and work with it.
        with tempfile.TemporaryDirectory() as temp_dir:
            SEPARATOR = '://'
            url = self._url

            # Prompt user input if required.
            if not self._user and self._user_key_exists:
                self._user = input(f'User ({url}): ')

            # Prompt password input if required.
            if not self._password and self._password_key_exists:
                self._password = getpass(f'Password ({url}): ', )

            url_parts = url.split(SEPARATOR)
            protocol = url_parts[0] if len(url_parts) > 1 else 'https'
            temp_url = url_parts[1] if len(url_parts) > 1 else url_parts[0]
            separator = SEPARATOR if protocol else ''
            user = self._user if self._user else ''
            password = self._password if self._password else ''
            colon = ':' if user and password else ''
            at = '@' if user or password else ''
            branch_command = f'--branch {self._branch}' if self._branch else ''
            url_with_credentials = f'{protocol}{separator}{user}{colon}{password}{at}{temp_url}'
            password = None

            code, _, stderr = execute_command(
                # Space before git-command to not log in history (might be disabled by the system).
                f' git clone {branch_command} --filter=blob:none --no-checkout {url_with_credentials} {temp_dir}',
            )

            # If clone was successful, go on.
            if code == 0:
                # Only clone desired target folder.
                code, _, stderr = execute_command(
                    f'git sparse-checkout set {self._target_path}' if self._target_path else '',
                    directory=temp_dir,
                )
            else:
                # Overwrite error to avoid reflecting the password from the URL to the output.
                stderr = ''

            if code == 0:
                # Only clone desired target folder.
                code, _, stderr = execute_command('git checkout', directory=temp_dir)

            # If checkout was successful, go on.
            if code == 0:
                # Make sure target directory exists.
                if self._target_path:
                    os.makedirs(join(temp_dir, self._target_path), exist_ok=True)
                target_file_path = join(self._target_path, file_name)
                target_file_path_full = join(temp_dir, target_file_path)

                # Write data to target file.
                with open(target_file_path_full, 'w') as f:
                    f.write(data)

                # Add changes.
                code, _, stderr = execute_command(f'git add "{target_file_path}"', directory=temp_dir)

                if code == 0:
                    # If necessary, request commit message.
                    commit_message = self._message if self._message else input('Commit message: ')

                    if not commit_message:
                        raise NoCommitMessageProvidedException(file_name)

                    # Commit changes.
                    code, _, stderr = execute_command(
                        f'git commit "{target_file_path}" -m "{commit_message}"',
                        directory=temp_dir,
                    )
                    no_changes = False

                    if code != 0:
                        if stderr.find('nothing to commit'):
                            # Everything's fine, the file just didn't change.
                            no_changes = True
                        else:
                            raise GitProblemException(f'{file_name} could not be committed to {url}', stderr)

                    if not no_changes:
                        # Push changes to repo (space before git-command to not log in history (might
                        # be disabled by the system)).
                        code, _, stderr = execute_command(
                            f' git push -u {url_with_credentials}',
                            directory=temp_dir,
                        )

                        if code != 0:
                            # Don't display Git error as it could reflect the password from the URL to the output.
                            raise GitProblemException(f'{file_name} could not be pushed to {url}')
                        
                else:
                    raise GitProblemException(f'{file_name} could not be added to Git')
            else:
                raise GitProblemException(f'Git repo {url} could not be cloned', stderr)
        return self



from pathlib import Path
from setuptools import setup, find_packages

# Setup process taken from here: https://www.freecodecamp.org/news/build-your-first-python-package/.

DESCRIPTION = 'Git distributor support for ninja-bear'
LONG_DESCRIPTION = Path(__file__).parent.absolute().joinpath('README.md').read_text()

setup(
    name='ninja-bear-distributor-git', 
    version='0.1.1',
    author='monstermichl',
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    py_modules=['ninja_bear_distributor_git'],
    entry_points = {
        'ninja_bear_distributor_git': ['config=ninja_bear_distributor_git.distributor:Distributor']
    },
    install_requires=[
    ],
    extras_require={
        'dev': [
            'ninja-bear>=0.1.4',
            'wheel>=0.41.1',
            'twine>=4.0.2',
            'ruff>=0.0.47',
            'coverage>=7.2.7',
            
        ],
    },
    python_requires='>=3.10',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    url='https://github.com/monstermichl/ninja-bear-distributor-git.git',
    keywords=[
        'ninja-bear',
        'plugin',
        'distributor',
        'git',
    ],
)
